#!/usr/bin/env python
# -*- coding: utf-8 -*-

from covsirphy.__version__ import __version__

__citation__ = f"""
CovsirPhy Development Team (2020-2022), CovsirPhy version {__version__}: Python library for COVID-19 analysis with phase-dependent SIR-derived ODE models, https://github.com/lisphilar/covid19-sir
Hirokazu Takaya (2020-2022), Kaggle Notebook, COVID-19 data with SIR model, https://www.kaggle.com/lisphilar/covid-19-data-with-sir-model
"""
